// Segmentation.h

#ifndef SEGMENTATION_H
#define SEGMENTATION_H

#include <vector>
#include <deque>
#include <list>
#include <queue>
#include <set>
#include <cmath>
#include "VoxIndex.h"
#include "SSLElem.h"
#include "Globals.h"

using std::vector;
using std::deque;
using std::list;
using std::priority_queue;
using std::set;
using std::multiset;

template <class T>
class reservable_priority_queue: public priority_queue<T>
{
public:
  typedef typename priority_queue<T>::size_type size_type;
  
  reservable_priority_queue(size_type capacity = 0) {
    reserve(capacity);
  }
  
  void reserve(size_type capacity) {
    this->c.reserve(capacity);
  } 
  size_type capacity() const {
    return this->c.capacity();
  } 
};

class Segmentation
{
public:
  Segmentation();
  ~Segmentation();
  
  static void growRegions();
  static void resetLabels();
  
  static void growRegionsS3DRG(void (*display)(void), short neighbors = 6);
  static void floodFillRegionS3DRG(short x, short y, short z, float tIntensity, void (*display)(void));
  
  static void growRegionsS2DRG(short neighbors = 4);
  
  static void SRG3D(vector<struct VoxIndex>, bool sslHeuristicSearch);
  
  static vector<vector<vector<short> > > getSegments() { return segments; }
  static vector<vector<vector<short> > > getSegmentSurfaces() { return segmentSurfaces; }
  static vector<struct VoxIndex> findSeeds();
  
  static unsigned short currentlySelectedSegment;
  static bool renderSegmentSurface;
  
private:
  static short neighborhood;
  static vector<vector<vector<short> > > segments;
  static vector<vector<vector<short> > > segmentSurfaces;
  static vector<vector<short> > currentRegion;
  static vector<vector<short> > currentRegionSurface;
  static unsigned short currentLabel;
  static const short MIN_REGION_SIZE;
  static void resetDataStructures();
  static void addVoxelToCurrentRegion(short x, short y, short z);
  static void addVoxelToCurrentRegionSurface(short x, short y, short z);
  static bool intensitiesAreEqual(float in1, float in2);
  static void floodFillRegion(short x, short y, short z, float tIntensity);
  
  static void floodFillRegionS2DRG(short x, short y, short z, float tIntensity);
  static bool floodFillRegionHelper(short x, short y, short z, vector<vector<short> >& queue,
                                    float& tIntensity);
  static bool floodFillRegionHelperS3DRG(short x, short y, short z, deque<vector<short> >& queue,
                                         float& tIntensity);
  static bool floodFillRegionHelperS2DRG(short x, short y, short z, vector<vector<short> >& queue,
                                         float& tIntensity);
  
  //static float getDelta(float intensity, float mean);
  //static bool validIndex(short x, short y, short z);
  
  static void add6NeighborsToSSL_DualHeap(struct VoxIndex vi,
                                          reservable_priority_queue<SSLElem>& ssl,
                                          reservable_priority_queue<SSLElem>& ssl2,
                                          float regionMean);
  /*
  static void getSSLMinElem(reservable_priority_queue<SSLElem>& ssl,
                            reservable_priority_queue<SSLElem>& ssl2,
                            SSLElem& resultElem);
  */
   
  static void add6NeighborsToSSL_Heap(struct VoxIndex vi, reservable_priority_queue<SSLElem>& ssl, float regionMean);
  static void add26NeighborsToSSL_Heap(struct VoxIndex vi, reservable_priority_queue<SSLElem>& ssl, float regionMean);
  static void addNeighborToSSL_Heap(short x, short y, short z, reservable_priority_queue<SSLElem>& ssl, float regionMean);
  static void insertSSLElem_Heap(SSLElem elem, reservable_priority_queue<SSLElem>& ssl);
  
  static void add6NeighborsToSSL_BinarySearchTree(struct VoxIndex voxel, multiset<SSLElem>& ssl, float regionMean);
  static void add26NeighborsToSSL_BinarySearchTree(struct VoxIndex vi, multiset<SSLElem>& ssl, float regionMean);
  static void addNeighborToSSL_BinarySearchTree(short x, short y, short z, multiset<SSLElem>& ssl, float regionMean);
  static void insertSSLElem_BinarySearchTree(SSLElem, multiset<SSLElem>& ssl);
  
  static void add6NeighborsToSSL_List(struct VoxIndex voxel, list<SSLElem>& ssl, float regionMean);
  static void add26NeighborsToSSL_List(struct VoxIndex vi, list<SSLElem>& ssl, float regionMean);
  static void addNeighborToSSL_List(short x, short y, short z, list<SSLElem>& ssl, float regionMean);
  static void insertSSLElem_List(SSLElem elem, list<SSLElem>& ssl);
  
  static void printSSL(multiset<SSLElem>& ssl);
  static void printSetInfo(set<int> s);
  static set<int> get6NeighborLabels(SSLElem& elem);
  static set<int> get26NeighborLabels(SSLElem& elem);
  static int getLabel(short x, short y, short z);
  
  static float sslDeltaCache;
  static float sslNumDeltasCache;
  
  static multiset<SSLElem>::iterator sslSearchHeuristic;
  static bool sslSearchHeuristicInitialized;
  static bool sslHeuristicSearchOn;
  
  static float getDelta(float intensity, float mean)
  {
    float p = 0.00009f;
    float diff = intensity - mean;
    float res = fabs(diff);
    //if (res < p) res = 0.0f;
    return res;
  }
  
  static bool validIndex(short x, short y, short z)
  {
    if (x >= 0 && x < GRID_SIZE && y >= 0 && y < GRID_SIZE &&
        z >= 0 && z < GRID_SIZE && voxels[x][y][z] != NULL) {
      return true;
    } else {
      return false;
    }
  }
  
  static void getSSLMinElem(reservable_priority_queue<SSLElem>& ssl,
                            reservable_priority_queue<SSLElem>& ssl2,
                            SSLElem& resultElem)
  {
    if (ssl.empty()) {
      resultElem = ssl2.top();
      ssl2.pop();
    } else if (ssl2.empty()) {
      resultElem = ssl.top();
      ssl.pop();
    } else {
      if (ssl.top().delta() < ssl2.top().delta()) {
        resultElem = ssl.top();
        ssl.pop();
      } else {
        resultElem = ssl2.top();
        ssl2.pop();
      }
    }
  }
};



#endif // SEGMENTATION_H